import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.utils import shuffle
import numpy as np
import pandas as pd
import re
from fuzzywuzzy import fuzz
import zipfile
import os
import random
from .text_functions import STOPWORDS, sentence_ngram, simple_textcleaning
from . import home
from .utils import download_file
from .skip_thought import train_model, batch_sequence

zip_location = home + '/rules-based.zip'

if not os.path.isfile(zip_location):
    print('downloading ZIP rules-based')
    download_file('rules-based.zip', zip_location)
    with zipfile.ZipFile(zip_location, 'r') as zip_ref:
        zip_ref.extractall(home)


def apply_stopwords_calon(string):
    string = re.sub('[^A-Za-z ]+', '', string)
    return ' '.join(
        [i for i in string.split() if i not in STOPWORDS and len(i) > 1]
    )


df = pd.read_csv(home + '/rules-based/calon.csv')
namacalon = df.NamaCalon.str.lower().unique().tolist()
for i in range(len(namacalon)):
    namacalon[i] = apply_stopwords_calon(namacalon[i])

df = pd.read_csv(home + '/rules-based/negeri.csv')
negeri = df.negeri.str.lower().unique().tolist()
parlimen = df.parlimen.str.lower().unique().tolist()
dun = df.dun.str.lower().unique().tolist()[:-1]

location = negeri + parlimen + dun

with open(home + '/rules-based/person-normalized', 'r') as fopen:
    person = list(filter(None, fopen.read().split('\n')))

person_dict = {}
for i in range(len(person)):
    splitted = person[i].split(':')
    uniques = list(
        filter(
            None,
            (
                set(
                    [k.strip().lower() for k in splitted[1].split(', ')]
                    + [splitted[0].lower()]
                )
            ),
        )
    )
    person_dict[splitted[0]] = uniques

with open(home + '/rules-based/topic-normalized', 'r') as fopen:
    topic = list(filter(None, fopen.read().split('\n')))

topic_dict = {}
for i in range(len(topic)):
    splitted = topic[i].split(':')
    uniques = list(
        filter(
            None,
            (
                set(
                    [k.strip().lower() for k in splitted[1].split(', ')]
                    + [splitted[0].lower()]
                )
            ),
        )
    )
    topic_dict[splitted[0]] = uniques

with open(home + '/rules-based/short-normalized', 'r') as fopen:
    short = list(filter(None, fopen.read().split('\n')))

short_dict = {}
for i in range(len(short)):
    splitted = short[i].split(':')
    uniques = list(
        filter(
            None,
            (
                set(
                    [k.strip().lower() for k in splitted[1].split(', ')]
                    + [splitted[0].lower()]
                )
            ),
        )
    )
    short_dict[splitted[0]] = uniques


def fuzzy_get_influencers(string):
    assert isinstance(string, str), 'input must be a string'
    string = string.lower()
    influencers = []
    for key, vals in person_dict.items():
        for v in vals:
            if fuzz.token_set_ratio(v, string) >= 80:
                influencers.append(key.lower())
                break
    for key, vals in short_dict.items():
        for v in vals:
            if v in string.split():
                influencers.append(key.lower())
                break

    for index in np.where(
        np.array([fuzz.token_set_ratio(i, string) for i in namacalon]) >= 80
    )[0]:
        influencers.append(namacalon[index].lower())
    return list(set(influencers))


def fuzzy_get_topics(string):
    assert isinstance(string, str), 'input must be a string'
    string = string.lower()
    topics = []
    for key, vals in topic_dict.items():
        for v in vals:
            if fuzz.token_set_ratio(v, string) >= 80:
                topics.append(key.lower())
                break
    for key, vals in person_dict.items():
        for v in vals:
            if fuzz.token_set_ratio(v, string) >= 80:
                topics.append(key.lower())
                break
    for key, vals in short_dict.items():
        for v in vals:
            if v in string.split():
                topics.append(key.lower())
                break

    return list(set(topics))


def is_location(string):
    for loc in location:
        if fuzz.token_set_ratio(loc.lower(), string) >= 90:
            return True
    return False


def fuzzy_get_location(string):
    assert isinstance(string, str), 'input must be a string'
    negeri_list = list(
        set([i for i in negeri if fuzz.token_set_ratio(i, string) >= 80])
    )
    parlimen_list = list(
        set([i for i in parlimen if fuzz.token_set_ratio(i, string) >= 80])
    )
    dun_list = list(
        set([i for i in dun if fuzz.token_set_ratio(i, string) >= 80])
    )
    return {'negeri': negeri_list, 'parlimen': parlimen_list, 'dun': dun_list}


class DEEP_SIMILARITY:
    def __init__(
        self,
        sess,
        model,
        vectorized,
        keys,
        dictionary,
        maxlen,
        is_influencers = False,
    ):
        self._sess = sess
        self._model = model
        self.vectorized = vectorized
        self.keys = keys
        self.dictionary = dictionary
        self.maxlen = maxlen
        self._is_influencers = is_influencers

    def get_similarity(self, string, anchor = 0.5):
        assert isinstance(string, str), 'input must be a string'
        assert isinstance(anchor, float), 'anchor must be a float'
        assert (
            anchor > 0 and anchor < 1
        ), 'anchor must be bigger than 0, less than 1'
        original_string = simple_textcleaning(string, decode = True)
        string = ' '.join(set(original_string.split()))
        encoded = self._sess.run(
            self._model.get_thought,
            feed_dict = {
                self._model.INPUT: batch_sequence(
                    [string], self.dictionary, maxlen = self.maxlen
                )
            },
        )
        where = np.where(
            cosine_similarity(self.vectorized, encoded)[:, 0] > anchor
        )[0]
        results = [self.keys[i].lower() for i in where]
        for key, vals in short_dict.items():
            for v in vals:
                if v in original_string.split():
                    results.append(key.lower())
                    break
        if self._is_influencers:
            for index in np.where(
                np.array(
                    [
                        fuzz.token_set_ratio(i, original_string)
                        for i in namacalon
                    ]
                )
                >= 80
            )[0]:
                results.append(namacalon[index].lower())
        return list(set(results))


class FAST_SIMILARITY:
    def __init__(self, vectorizer, vectorized, keys, is_influencers = False):
        self.vectorizer = vectorizer
        self.vectorized = vectorized
        self.keys = keys
        self._is_influencers = is_influencers

    def get_similarity(self, string, anchor = 0.1):
        assert isinstance(string, str), 'input must be a string'
        assert isinstance(anchor, float), 'anchor must be a float'
        assert (
            anchor > 0 and anchor < 1
        ), 'anchor must be bigger than 0, less than 1'
        original_string = simple_textcleaning(string, decode = True)
        string = ' '.join(set(original_string.split()))
        where = np.where(
            cosine_similarity(
                self.vectorized, self.vectorizer.transform([string])
            )[:, 0]
            > anchor
        )[0]
        results = [self.keys[i].lower() for i in where]
        for key, vals in short_dict.items():
            for v in vals:
                if v in original_string.split():
                    results.append(key.lower())
                    break
        if self._is_influencers:
            for index in np.where(
                np.array(
                    [
                        fuzz.token_set_ratio(i, original_string)
                        for i in namacalon
                    ]
                )
                >= 80
            )[0]:
                results.append(namacalon[index].lower())
        return list(set(results))


def generate_topics():
    texts = [' '.join(words) for _, words in topic_dict.items()]
    keys = [key for key, _ in topic_dict.items()]
    texts += [' '.join(words) for _, words in person_dict.items()]
    keys += [key for key, _ in person_dict.items()]
    texts = [' '.join(list(set(text.split()))) for text in texts]
    output = []
    for text in texts:
        output.append(
            ' '.join([word for word in text.split() if word not in STOPWORDS])
        )
    return output, keys


def generate_influencers():
    texts = [' '.join(words) for _, words in person_dict.items()]
    keys = [key for key, _ in person_dict.items()]
    texts = [' '.join(list(set(text.split()))) for text in texts]
    output = []
    for text in texts:
        output.append(
            ' '.join([word for word in text.split() if word not in STOPWORDS])
        )
    return output, keys


def deep_get_topics(
    epoch = 5,
    batch_size = 16,
    embedding_size = 256,
    output_size = 300,
    maxlen = 100,
    ngrams = (1, 4),
):
    assert isinstance(epoch, int), 'epoch must be an integer'
    assert isinstance(batch_size, int), 'batch_size must be an integer'
    assert isinstance(embedding_size, int), 'embedding_size must be an integer'
    assert isinstance(output_size, int), 'output_size must be an integer'
    assert isinstance(maxlen, int), 'maxlen must be an integer'
    assert isinstance(ngrams, tuple), 'ngrams must be a tuple'
    output, keys = generate_topics()
    batch_x, batch_y = [], []
    for i in range(len(output)):
        augmentation = sentence_ngram(output[i])
        batch_y.extend([keys[i]] * len(augmentation))
        batch_x.extend(augmentation)
    batch_x, batch_y = shuffle(batch_x, batch_y)
    sess, model, dictionary = train_model(
        batch_x,
        batch_y,
        epoch = epoch,
        batch_size = batch_size,
        embedding_size = embedding_size,
        output_size = output_size,
        maxlen = maxlen,
    )
    encoded = sess.run(
        model.get_thought,
        feed_dict = {
            model.INPUT: batch_sequence(output, dictionary, maxlen = maxlen)
        },
    )
    return DEEP_SIMILARITY(sess, model, encoded, keys, dictionary, maxlen)


def deep_get_influencers(
    epoch = 10,
    batch_size = 16,
    embedding_size = 256,
    output_size = 300,
    maxlen = 100,
    ngrams = (1, 4),
):
    assert isinstance(epoch, int), 'epoch must be an integer'
    assert isinstance(batch_size, int), 'batch_size must be an integer'
    assert isinstance(embedding_size, int), 'embedding_size must be an integer'
    assert isinstance(output_size, int), 'output_size must be an integer'
    assert isinstance(maxlen, int), 'maxlen must be an integer'
    assert isinstance(ngrams, tuple), 'ngrams must be a tuple'
    output, keys = generate_influencers()
    batch_x, batch_y = [], []
    for i in range(len(output)):
        augmentation = sentence_ngram(output[i])
        batch_y.extend([keys[i]] * len(augmentation))
        batch_x.extend(augmentation)
    assert batch_size < len(batch_x), 'batch size must smaller with corpus size'
    batch_x, batch_y = shuffle(batch_x, batch_y)
    sess, model, dictionary = train_model(
        batch_x,
        batch_y,
        epoch = epoch,
        batch_size = batch_size,
        embedding_size = embedding_size,
        output_size = output_size,
        maxlen = maxlen,
    )
    encoded = sess.run(
        model.get_thought,
        feed_dict = {
            model.INPUT: batch_sequence(output, dictionary, maxlen = maxlen)
        },
    )
    return DEEP_SIMILARITY(
        sess, model, encoded, keys, dictionary, maxlen, is_influencers = True
    )


def fast_get_topics(vectorizer = 'tfidf', ngrams = (3, 10)):
    assert isinstance(vectorizer, str), 'vectorizer must be a string'
    assert isinstance(ngrams, tuple), 'ngrams must be a tuple'
    if 'tfidf' in vectorizer.lower():
        char_vectorizer = TfidfVectorizer(
            sublinear_tf = True,
            strip_accents = 'unicode',
            analyzer = 'char',
            ngram_range = ngrams,
        )
    elif 'count' in vectorizer.lower():
        char_vectorizer = CountVectorizer(
            strip_accents = 'unicode', analyzer = 'char', ngram_range = ngrams
        )
    else:
        raise Exception('model not supported')
    output, keys = generate_topics()
    vectorizer = char_vectorizer.fit(output)
    vectorized = vectorizer.transform(output)
    return FAST_SIMILARITY(vectorizer, vectorized, keys)


def fast_get_influencers(vectorizer = 'tfidf', ngrams = (3, 10)):
    assert isinstance(vectorizer, str), 'vectorizer must be a string'
    assert isinstance(ngrams, tuple), 'ngrams must be a tuple'
    if 'tfidf' in vectorizer.lower():
        char_vectorizer = TfidfVectorizer(
            sublinear_tf = True,
            strip_accents = 'unicode',
            analyzer = 'char',
            ngram_range = ngrams,
        )
    elif 'count' in vectorizer.lower():
        char_vectorizer = CountVectorizer(
            strip_accents = 'unicode', analyzer = 'char', ngram_range = ngrams
        )
    else:
        raise Exception('model not supported')
    output, keys = generate_influencers()
    vectorizer = char_vectorizer.fit(output)
    vectorized = vectorizer.transform(output)
    return FAST_SIMILARITY(vectorizer, vectorized, keys, is_influencers = True)



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import pickle
import os
import xgboost as xgb
import numpy as np
from .text_functions import simple_textcleaning
from .utils import download_file
from . import home

bow_pkl = home + '/bow-language-detection.pkl'
multinomial_pkl = home + '/multinomial-language-detection.pkl'

xgb_bow_pkl = home + '/bow-xgb-language-detection.pkl'
xgb_pkl = home + '/xgb-language-detection.pkl'

lang_labels = {0: 'OTHER', 1: 'ENGLISH', 2: 'INDONESIA', 3: 'MALAY'}


def get_language_labels():
    return lang_labels


class USER_XGB:
    def __init__(self, xgb, label, vectorize):
        self.xgb = xgb
        self.label = label
        self.vectorize = vectorize

    def predict(self, string, get_proba = False):
        assert isinstance(string, str), 'input must be a string'
        vectors = self.vectorize.transform([simple_textcleaning(string, True)])
        result = self.xgb.predict(
            xgb.DMatrix(vectors), ntree_limit = self.xgb.best_ntree_limit
        )[0]
        if get_proba:
            return {self.label[i]: result[i] for i in range(len(result))}
        else:
            return self.label[np.argmax(result)]

    def predict_batch(self, strings, get_proba = False):
        assert isinstance(strings, list) and isinstance(
            strings[0], str
        ), 'input must be list of strings'
        strings = [simple_textcleaning(string, True) for string in strings]
        vectors = self.vectorize.transform(strings)
        results = self.xgb.predict(
            xgb.DMatrix(vectors), ntree_limit = self.xgb.best_ntree_limit
        )
        if get_proba:
            outputs = []
            for result in results:
                outputs.append(
                    {self.label[i]: result[i] for i in range(len(result))}
                )
            return outputs
        else:
            return [self.label[i] for i in np.argmax(results, axis = 1)]


class USER_BAYES:
    def __init__(self, multinomial, label, vectorize):
        self.multinomial = multinomial
        self.label = label
        self.vectorize = vectorize

    def predict(self, string, get_proba = False):
        assert isinstance(string, str), 'input must be a string'
        vectors = self.vectorize.transform([simple_textcleaning(string, True)])
        if get_proba:
            result = self.multinomial.predict_proba(vectors)[0]
            return {self.label[i]: result[i] for i in range(len(result))}
        else:
            return self.label[self.multinomial.predict(vectors)[0]]

    def predict_batch(self, strings, get_proba = False):
        assert isinstance(strings, list) and isinstance(
            strings[0], str
        ), 'input must be list of strings'
        strings = [simple_textcleaning(string, True) for string in strings]
        vectors = self.vectorize.transform(strings)
        if get_proba:
            results = self.multinomial.predict_proba(vectors)
            outputs = []
            for result in results:
                outputs.append(
                    {self.label[i]: result[i] for i in range(len(result))}
                )
            return outputs
        else:
            return [
                self.label[result]
                for result in self.multinomial.predict(vectors)
            ]


def multinomial_detect_languages():
    if not os.path.isfile(bow_pkl):
        print('downloading LANGUAGE-DETECTION pickled bag-of-word')
        download_file('bow-language-detection.pkl', bow_pkl)
    if not os.path.isfile(multinomial_pkl):
        print('downloading LANGUAGE-DETECTION pickled multinomial model')
        download_file('multinomial-language-detection.pkl', multinomial_pkl)
    with open(bow_pkl, 'rb') as fopen:
        BOW = pickle.load(fopen)
    with open(multinomial_pkl, 'rb') as fopen:
        MULTINOMIAL = pickle.load(fopen)
    return USER_BAYES(MULTINOMIAL, lang_labels, BOW)


def xgb_detect_languages():
    if not os.path.isfile(xgb_bow_pkl):
        print('downloading LANGUAGE-DETECTION pickled bag-of-word XGB')
        download_file('bow-xgboost-language-detection.pkl', xgb_bow_pkl)
    if not os.path.isfile(xgb_pkl):
        print('downloading LANGUAGE-DETECTION pickled XGB model')
        download_file('xgboost-language-detection.pkl', xgb_pkl)
    with open(xgb_bow_pkl, 'rb') as fopen:
        BOW = pickle.load(fopen)
    with open(xgb_pkl, 'rb') as fopen:
        XGB = pickle.load(fopen)
    return USER_XGB(XGB, lang_labels, BOW)



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from sklearn.utils import shuffle
from sklearn.manifold import TSNE
from scipy.spatial.distance import cdist
from sklearn.neighbors import NearestNeighbors
from fuzzywuzzy import fuzz
import pickle
import os
import sys
import collections
import re
import numpy as np
import tensorflow as tf
from . import home
from .utils import download_file


def malaya_word2vec(size = 256):
    assert isinstance(size, int), 'input must be an integer'
    if size not in [32, 64, 128, 256, 512]:
        raise Exception('size word2vec not supported')
    if not os.path.isfile('%s/word2vec-%d.p' % (home, size)):
        print('downloading word2vec-%d embedded' % (size))
        download_file(
            'v7/word2vec/word2vec-%d.p' % (size),
            '%s/word2vec-%d.p' % (home, size),
        )
    with open('%s/word2vec-%d.p' % (home, size), 'rb') as fopen:
        return pickle.load(fopen)


class Calculator:
    def __init__(self, tokens):
        self._tokens = tokens
        self._current = tokens[0]

    def exp(self):
        result = self.term()
        while self._current in ('+', '-'):
            if self._current == '+':
                self.next()
                result += self.term()
            if self._current == '-':
                self.next()
                result -= self.term()
        return result

    def factor(self):
        result = None
        if self._current[0].isdigit() or self._current[-1].isdigit():
            result = np.array([float(i) for i in self._current.split(',')])
            self.next()
        elif self._current is '(':
            self.next()
            result = self.exp()
            self.next()
        return result

    def next(self):
        self._tokens = self._tokens[1:]
        self._current = self._tokens[0] if len(self._tokens) > 0 else None

    def term(self):
        result = self.factor()
        while self._current in ('*', '/'):
            if self._current == '*':
                self.next()
                result *= self.term()
            if self._current == '/':
                self.next()
                result /= self.term()
        return result


class Word2Vec:
    def __init__(self, embed_matrix, dictionary):
        self._embed_matrix = embed_matrix
        self._dictionary = dictionary
        self._reverse_dictionary = {v: k for k, v in dictionary.items()}
        self.words = list(dictionary.keys())

    def get_vector_by_name(self, word):
        return np.ravel(self._embed_matrix[self._dictionary[word], :])

    def calculator(
        self,
        equation,
        num_closest = 5,
        metric = 'cosine',
        return_similarity = True,
    ):
        assert isinstance(equation, str), 'input must be a string'
        assert isinstance(num_closest, int), 'num_closest must be an integer'
        assert isinstance(metric, str), 'metric must be a string'
        assert isinstance(
            return_similarity, bool
        ), 'num_closest must be a boolean'
        tokens, temp = [], ''
        for char in equation:
            if char == ' ':
                continue
            if char not in '()*+-':
                temp += char
            else:
                if len(temp):
                    row = self._dictionary[
                        self.words[
                            np.argmax([fuzz.ratio(temp, k) for k in self.words])
                        ]
                    ]
                    tokens.append(
                        ','.join(
                            self._embed_matrix[row, :].astype('str').tolist()
                        )
                    )
                    temp = ''
                tokens.append(char)
        if len(temp):
            row = self._dictionary[
                self.words[np.argmax([fuzz.ratio(temp, k) for k in self.words])]
            ]
            tokens.append(
                ','.join(self._embed_matrix[row, :].astype('str').tolist())
            )
        if return_similarity:
            nn = NearestNeighbors(num_closest + 1, metric = metric).fit(
                self._embed_matrix
            )
            distances, idx = nn.kneighbors(
                Calculator(tokens).exp().reshape((1, -1))
            )
            word_list = []
            for i in range(1, idx.shape[1]):
                word_list.append(
                    [self._reverse_dictionary[idx[0, i]], 1 - distances[0, i]]
                )
            return word_list
        else:
            closest_indices = self.closest_row_indices(
                Calculator(tokens).exp(), num_closest + 1, metric
            )
            word_list = []
            for i in closest_indices:
                word_list.append(self._reverse_dictionary[i])
            return word_list

    def n_closest(
        self, word, num_closest = 5, metric = 'cosine', return_similarity = True
    ):
        assert isinstance(word, str), 'input must be a string'
        assert isinstance(num_closest, int), 'num_closest must be an integer'
        assert isinstance(metric, str), 'metric must be a string'
        assert isinstance(
            return_similarity, bool
        ), 'num_closest must be a boolean'
        if return_similarity:
            nn = NearestNeighbors(num_closest + 1, metric = metric).fit(
                self._embed_matrix
            )
            distances, idx = nn.kneighbors(
                self._embed_matrix[self._dictionary[word], :].reshape((1, -1))
            )
            word_list = []
            for i in range(1, idx.shape[1]):
                word_list.append(
                    [self._reverse_dictionary[idx[0, i]], 1 - distances[0, i]]
                )
            return word_list
        else:
            wv = self.get_vector_by_name(word)
            closest_indices = self.closest_row_indices(
                wv, num_closest + 1, metric
            )
            word_list = []
            for i in closest_indices:
                word_list.append(self._reverse_dictionary[i])
            if word in word_list:
                word_list.remove(word)
            return word_list

    def closest_row_indices(self, wv, num, metric):
        dist_array = np.ravel(
            cdist(self._embed_matrix, wv.reshape((1, -1)), metric = metric)
        )
        sorted_indices = np.argsort(dist_array)
        return sorted_indices[:num]

    def analogy(self, a, b, c, num = 1, metric = 'cosine'):
        assert isinstance(a, str), 'a must be a string'
        assert isinstance(b, str), 'b must be a string'
        assert isinstance(c, str), 'c must be a string'
        if a not in self._dictionary:
            raise Exception('a not in dictinary')
        if b not in self._dictionary:
            raise Exception('b not in dictinary')
        if c not in self._dictionary:
            raise Exception('c not in dictinary')
        va = self.get_vector_by_name(a)
        vb = self.get_vector_by_name(b)
        vc = self.get_vector_by_name(c)
        vd = vb - va + vc
        closest_indices = self.closest_row_indices(vd, num, metric)
        d_word_list = []
        for i in closest_indices:
            d_word_list.append(self._reverse_dictionary[i])
        return d_word_list

    def project_2d(self, start, end):
        assert isinstance(start, int), 'start must be an integer'
        assert isinstance(end, int), 'end must be an integer'
        tsne = TSNE(n_components = 2)
        embed_2d = tsne.fit_transform(self._embed_matrix[start:end, :])
        word_list = []
        for i in range(start, end):
            word_list.append(self._reverse_dictionary[i])
        return embed_2d, word_list



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from nltk.tokenize import word_tokenize
import os
import re
import pickle
import json
import xgboost as xgb
import numpy as np
from . import home
from .tatabahasa import tatabahasa_dict, hujung, permulaan
from .utils import download_file
from .text_functions import entities_textcleaning
from .keras_model import (
    get_char_bidirectional,
    get_crf_lstm_bidirectional,
    get_crf_lstm_concat_bidirectional,
)
from .keras_model import CONCAT_MODEL, WORD_MODEL, CHAR_MODEL

bow_pkl = home + '/bow-pos.pkl'
multinomial_pkl = home + '/multinomial-pos.pkl'
xgb_bow_pkl = home + '/xgb-bow-pos.pkl'
xgb_pkl = home + '/xgb-pos.pkl'

MULTINOMIAL, BOW = None, None
XGB, BOW_XGB = None, None

pos_labels = {
    0: 'ADJ',
    1: 'ADP',
    2: 'ADV',
    3: 'AUX',
    4: 'CCONJ',
    5: 'DET',
    6: 'NOUN',
    7: 'NUM',
    8: 'PART',
    9: 'PRON',
    10: 'PROPN',
    11: 'SCONJ',
    12: 'SYM',
    13: 'VERB',
    14: 'X',
}


def naive_POS_word(word):
    for key, vals in tatabahasa_dict.items():
        if word in vals:
            return (key, word)
    try:
        if len(re.findall(r'^(.*?)(%s)$' % ('|'.join(hujung[:1])), i)[0]) > 1:
            return ('KJ', word)
    except:
        pass
    try:
        if (
            len(re.findall(r'^(.*?)(%s)' % ('|'.join(permulaan[:-4])), word)[0])
            > 1
        ):
            return ('KJ', word)
    except Exception as e:
        pass
    if len(word) > 2:
        return ('KN', word)
    else:
        return ('', word)


def naive_pos(string):
    assert isinstance(string, str), 'input must be a string'
    string = string.lower()
    results = []
    for i in word_tokenize(string):
        results.append(naive_POS_word(i))
    return results


def multinomial_pos(string):
    assert isinstance(string, str), 'input must be a string'
    global MULTINOMIAL, BOW
    string = entities_textcleaning(string)
    if MULTINOMIAL is None and BOW is None:
        if not os.path.isfile(bow_pkl):
            print('downloading pickled bag-of-word multinomial POS')
            download_file('bow-entities.pkl', bow_pkl)
        if not os.path.isfile(multinomial_pkl):
            print('downloading pickled multinomial POS model')
            download_file('multinomial-entities.pkl', multinomial_pkl)
        with open(bow_pkl, 'rb') as fopen:
            BOW = pickle.load(fopen)
        with open(multinomial_pkl, 'rb') as fopen:
            MULTINOMIAL = pickle.load(fopen)
    return [
        (string[no], pos_labels[i])
        for no, i in enumerate(MULTINOMIAL.predict(BOW.transform(string)))
    ]


def xgb_pos(string):
    assert isinstance(string, str), 'input must be a string'
    global XGB, BOW_XGB
    string = entities_textcleaning(string)
    if XGB is None and BOW_XGB is None:
        if not os.path.isfile(xgb_bow_pkl):
            print('downloading pickled bag-of-word XGB POS')
            download_file('xgb-bow-pos.pkl', xgb_bow_pkl)
        if not os.path.isfile(xgb_pkl):
            print('downloading pickled xgb POS model')
            download_file('xgb-pos.pkl', xgb_pkl)
        with open(xgb_bow_pkl, 'rb') as fopen:
            BOW_XGB = pickle.load(fopen)
        with open(xgb_pkl, 'rb') as fopen:
            XGB = pickle.load(fopen)
    dmatrix = xgb.DMatrix(BOW_XGB.transform(string))
    results = np.argmax(
        XGB.predict(dmatrix, ntree_limit = XGB.best_ntree_limit), axis = 1
    )
    return [(string[no], pos_labels[i]) for no, i in enumerate(results)]


def get_available_pos_models():
    return ['char', 'word', 'concat']


char_frozen = home + '/pos-char-frozen.h5'
word_frozen = home + '/pos-word-frozen.h5'
concat_frozen = home + '/pos-concat-frozen.h5'

char_settings = home + '/pos-char-settings.json'
word_settings = home + '/pos-word-settings.json'
concat_settings = home + '/pos-concat-settings.json'

char_frozen_location = 'v7/pos/char-bidirectional-pos.h5'
char_settings_location = 'v7/pos/char-bidirectional-pos.json'

word_frozen_location = 'v7/pos/crf-lstm-bidirectional-pos.h5'
word_settings_location = 'v7/pos/crf-lstm-bidirectional-pos.json'

concat_frozen_location = 'v7/pos/crf-lstm-concat-bidirectional-pos.h5'
concat_settings_location = 'v7/pos/crf-lstm-concat-bidirectional-pos.json'


def deep_pos(model = 'concat'):
    assert isinstance(model, str), 'model must be a string'
    if model == 'char':
        if not os.path.isfile(char_settings):
            print('downloading POS char settings')
            download_file(char_settings_location, char_settings)
        with open(char_settings, 'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(char_frozen):
            print('downloading POS char frozen model')
            download_file(char_frozen_location, char_frozen)
        model = get_char_bidirectional(nodes['char2idx'], nodes['tag2idx'])
        model.load_weights(char_frozen)
        return CHAR_MODEL(model, nodes, is_lower = False)
    elif model == 'word':
        if not os.path.isfile(word_settings):
            print('downloading POS word settings')
            download_file(word_settings_location, word_settings)
        with open(word_settings, 'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(word_frozen):
            print('downloading POS word frozen model')
            download_file(word_frozen_location, word_frozen)
        model = get_crf_lstm_bidirectional(nodes['word2idx'], nodes['tag2idx'])
        model.load_weights(word_frozen)
        return WORD_MODEL(model, nodes, is_lower = False)
    elif model == 'concat':
        if not os.path.isfile(concat_settings):
            print('downloading POS concat settings')
            download_file(concat_settings_location, concat_settings)
        with open(concat_settings, 'r') as fopen:
            nodes = json.loads(fopen.read())
            nodes['idx2word'] = {
                int(k): v for k, v in nodes['idx2word'].items()
            }
        if not os.path.isfile(concat_frozen):
            print('downloading POS concat frozen model')
            download_file(concat_frozen_location, concat_frozen)
        model = get_crf_lstm_concat_bidirectional(
            nodes['char2idx'], nodes['word2idx'], nodes['tag2idx']
        )
        model.load_weights(concat_frozen)
        return CONCAT_MODEL(model, nodes, is_lower = False)
    else:
        raise Exception('model not supported')



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.decomposition import TruncatedSVD, NMF, LatentDirichletAllocation
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import Normalizer
from unidecode import unidecode
from .stemmer import sastrawi_stemmer
import itertools
import numpy as np
import re
from .text_functions import (
    simple_textcleaning,
    STOPWORDS,
    classification_textcleaning,
    print_topics_modelling,
)
from .lda2vec import process_data, generate_lda, train_lda2vec


class DEEP_TOPIC:
    def __init__(self, idx2word, similarity, corpus, doc_weights):
        self.idx2word = idx2word
        self.similarity = similarity
        self.corpus = corpus
        self.doc_weights = doc_weights

    def print_topics(self, len_topic):
        assert isinstance(len_topic, int), 'len_topic must be an integer'
        print_topics_modelling(
            range(len_topic),
            feature_names = np.array(self.idx2word),
            sorting = np.argsort(self.similarity)[:, ::-1],
            topics_per_chunk = 5,
            n_words = 10,
        )

    def get_topics(self, len_topic):
        assert isinstance(len_topic, int), 'len_topic must be an integer'
        results = []
        for no, topic in enumerate(self.similarity):
            results.append(
                (
                    no,
                    ' '.join(
                        [
                            self.idx2word[i]
                            for i in topic.argsort()[: -len_topic - 1 : -1]
                        ]
                    ),
                )
            )
        return results

    def get_sentences(self, len_sentence, k = 0):
        assert isinstance(len_sentence, int), 'len_sentence must be an integer'
        assert isinstance(k, int), 'k must be an integer'
        reverse_sorted = np.argsort(self.doc_weights[:, k])[::-1]
        return [self.corpus[i] for i in reverse_sorted[:len_sentence]]


class TOPIC:
    def __init__(self, features, comp, corpus, transformed):
        self.features = features
        self.comp = comp
        self.corpus = corpus
        self.transformed = transformed

    def print_topics(self, len_topic):
        assert isinstance(len_topic, int), 'len_topic must be an integer'
        print_topics_modelling(
            range(len_topic),
            feature_names = np.array(self.features),
            sorting = np.argsort(self.comp.components_)[:, ::-1],
            topics_per_chunk = 5,
            n_words = 10,
        )

    def get_topics(self, len_topic):
        assert isinstance(len_topic, int), 'len_topic must be an integer'
        results = []
        for no, topic in enumerate(self.comp.components_):
            results.append(
                (
                    no,
                    ' '.join(
                        [
                            self.features[i]
                            for i in topic.argsort()[: -len_topic - 1 : -1]
                        ]
                    ),
                )
            )
        return results

    def get_sentences(self, len_sentence, k = 0):
        assert isinstance(len_sentence, int), 'len_sentence must be an integer'
        assert isinstance(k, int), 'k must be an integer'
        reverse_sorted = np.argsort(self.transformed[:, k])[::-1]
        return [self.corpus[i] for i in reverse_sorted[:len_sentence]]


def lda_topic_modelling(
    corpus,
    n_topics = 10,
    max_df = 0.95,
    min_df = 2,
    stemming = True,
    cleaning = simple_textcleaning,
    stop_words = STOPWORDS,
):
    assert isinstance(corpus, list) and isinstance(
        corpus[0], str
    ), 'input must be list of strings'
    assert isinstance(n_topics, int), 'n_topics must be an integer'
    assert isinstance(max_df, float), 'max_df must be a float'
    assert isinstance(min_df, int), 'min_df must be an integer'
    assert isinstance(stemming, bool), 'bool must be a boolean'
    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi_stemmer(corpus[i])
    tf_vectorizer = CountVectorizer(
        max_df = max_df, min_df = min_df, stop_words = stop_words
    )
    tf = tf_vectorizer.fit_transform(corpus)
    tf_features = tf_vectorizer.get_feature_names()
    lda = LatentDirichletAllocation(
        n_topics = n_topics,
        max_iter = 5,
        learning_method = 'online',
        learning_offset = 50.0,
        random_state = 0,
    ).fit(tf)
    return TOPIC(
        tf_features,
        lda,
        [classification_textcleaning(c) for c in corpus],
        lda.transform(tf),
    )


def nmf_topic_modelling(
    corpus,
    n_topics = 10,
    max_df = 0.95,
    min_df = 2,
    stemming = True,
    cleaning = simple_textcleaning,
    stop_words = STOPWORDS,
):
    assert isinstance(corpus, list) and isinstance(
        corpus[0], str
    ), 'input must be list of strings'
    assert isinstance(n_topics, int), 'n_topics must be an integer'
    assert isinstance(max_df, float), 'max_df must be a float'
    assert isinstance(min_df, int), 'min_df must be an integer'
    assert isinstance(stemming, bool), 'bool must be a boolean'
    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi_stemmer(corpus[i])
    tfidf_vectorizer = TfidfVectorizer(
        max_df = max_df, min_df = min_df, stop_words = stop_words
    )
    tfidf = tfidf_vectorizer.fit_transform(corpus)
    tfidf_features = tfidf_vectorizer.get_feature_names()
    nmf = NMF(
        n_components = n_topics,
        random_state = 1,
        alpha = 0.1,
        l1_ratio = 0.5,
        init = 'nndsvd',
    ).fit(tfidf)
    return TOPIC(
        tfidf_features,
        nmf,
        [classification_textcleaning(c) for c in corpus],
        nmf.transform(tfidf),
    )


def lsa_topic_modelling(
    corpus,
    n_topics,
    max_df = 0.95,
    min_df = 2,
    stemming = True,
    cleaning = simple_textcleaning,
    stop_words = STOPWORDS,
):
    assert isinstance(corpus, list) and isinstance(
        corpus[0], str
    ), 'input must be list of strings'
    assert isinstance(n_topics, int), 'n_topics must be an integer'
    assert isinstance(max_df, float), 'max_df must be a float'
    assert isinstance(min_df, int), 'min_df must be an integer'
    assert isinstance(stemming, bool), 'bool must be a boolean'
    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi_stemmer(corpus[i])
    tfidf_vectorizer = TfidfVectorizer(
        max_df = max_df, min_df = min_df, stop_words = stop_words
    )
    tfidf = tfidf_vectorizer.fit_transform(corpus)
    tfidf_features = tfidf_vectorizer.get_feature_names()
    tfidf = Normalizer().fit_transform(tfidf)
    lsa = TruncatedSVD(n_topics).fit(tfidf)
    return TOPIC(
        tfidf_features,
        lsa,
        [classification_textcleaning(c) for c in corpus],
        lsa.transform(tfidf),
    )


def lda2vec_topic_modelling(
    corpus,
    n_topics,
    word_vectors,
    word2idx,
    idx2word,
    max_df = 0.95,
    min_df = 2,
    stemming = True,
    cleaning = simple_textcleaning,
    min_words = 5,
    epoch = 10,
    batch_size = 32,
):
    assert isinstance(corpus, list) and isinstance(
        corpus[0], str
    ), 'input must be list of strings'
    assert batch_size < len(corpus), 'batch size must smaller with corpus size'
    if cleaning is not None:
        for i in range(len(corpus)):
            corpus[i] = cleaning(corpus[i])
    if stemming:
        for i in range(len(corpus)):
            corpus[i] = sastrawi_stemmer(corpus[i])
    encoded_data, unigram_distribution = process_data(
        corpus, word2idx, min_words = min_words
    )
    doc_weights_init = generate_lda(
        corpus, n_topics, max_df = max_df, min_df = min_df
    )
    doc_weights, similarity = train_lda2vec(
        encoded_data,
        unigram_distribution,
        word_vectors,
        doc_weights_init,
        n_topics,
        batch_size = batch_size,
        epoch = epoch,
    )
    return DEEP_TOPIC(idx2word, similarity, corpus, doc_weights)



import os
from distutils.core import setup
from distutils import dir_util
from distutils import sysconfig
import setuptools
import atexit
from setuptools.command.install import install


def _post_install():
    print(
        'Please install keras-contrib, `pip install git+https://www.github.com/keras-team/keras-contrib.git`'
    )
    print(
        'Cudnn 7 and above seems has problem with Malaya fast-text models, we prefer to use Tensorflow Version 1.5 CUDA 8.0 Cudnn 5'
    )
    print(
        'You can simply downgrade by `pip uninstall tensorflow && pip install tensorflow==1.5`'
    )


class new_install(install):
    def __init__(self, *args, **kwargs):
        super(new_install, self).__init__(*args, **kwargs)
        atexit.register(_post_install)


__packagename__ = 'malaya'

setuptools.setup(
    name = __packagename__,
    packages = setuptools.find_packages(),
    version = '0.7.1',
    description = 'Natural-Language-Toolkit for bahasa Malaysia, powered by Deep Learning.',
    author = 'huseinzol05',
    author_email = 'husein.zol05@gmail.com',
    url = 'https://github.com/DevconX/Malaya',
    download_url = 'https://github.com/DevconX/Malaya/archive/master.zip',
    keywords = ['nlp', 'bm'],
    install_requires = [
        'xgboost==0.80',
        'sklearn',
        'scikit-learn==0.19.1',
        'requests',
        'fuzzywuzzy',
        'tqdm',
        'nltk',
        'unidecode',
        'tensorflow',
        'numpy',
        'scipy',
        'python-levenshtein',
        'pandas',
        'keras',
        'PySastrawi',
    ],
    cmdclass = {'install': new_install},
    classifiers = [
        'Programming Language :: Python :: 3',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Text Processing',
    ],
    long_description = open('readme-pypi.srt').read(),
)



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import pickle
import os
import json
import xgboost as xgb
import numpy as np
from .utils import download_file
from . import home
from .text_functions import entities_textcleaning
from .keras_model import (
    get_char_bidirectional,
    get_crf_lstm_bidirectional,
    get_crf_lstm_concat_bidirectional,
)
from .keras_model import CONCAT_MODEL, WORD_MODEL, CHAR_MODEL

bow_pkl = home + '/bow-entities.pkl'
multinomial_pkl = home + '/multinomial-entities.pkl'
xgb_bow_pkl = home + '/xgb-bow-entities.pkl'
xgb_pkl = home + '/xgb-entities.pkl'

MULTINOMIAL, BOW = None, None
XGB, BOW_XGB = None, None

entities_labels = {
    0: 'OTHER',
    1: 'law',
    2: 'location',
    3: 'organization',
    4: 'person',
    5: 'quantity',
    6: 'time',
}


def multinomial_entities(string):
    assert isinstance(string, str), 'input must be a string'
    global MULTINOMIAL, BOW
    string = entities_textcleaning(string)
    if MULTINOMIAL is None and BOW is None:
        if not os.path.isfile(bow_pkl):
            print('downloading ENTITIES pickled bag-of-word multinomial')
            download_file('bow-entities.pkl', bow_pkl)
        if not os.path.isfile(multinomial_pkl):
            print('downloading ENTITIES pickled multinomial model')
            download_file('multinomial-entities.pkl', multinomial_pkl)
        with open(bow_pkl, 'rb') as fopen:
            BOW = pickle.load(fopen)
        with open(multinomial_pkl, 'rb') as fopen:
            MULTINOMIAL = pickle.load(fopen)
    return [
        (string[no], entities_labels[i])
        for no, i in enumerate(MULTINOMIAL.predict(BOW.transform(string)))
    ]


def xgb_entities(string):
    assert isinstance(string, str), 'input must be a string'
    global XGB, BOW_XGB
    string = entities_textcleaning(string)
    if XGB is None and BOW_XGB is None:
        if not os.path.isfile(xgb_bow_pkl):
            print('downloading ENTITIES pickled bag-of-word XGB')
            download_file('xgb-bow-entities.pkl', xgb_bow_pkl)
        if not os.path.isfile(xgb_pkl):
            print('downloading ENTITIES pickled XGB model')
            download_file('xgb-entities.pkl', xgb_pkl)
        with open(xgb_bow_pkl, 'rb') as fopen:
            BOW_XGB = pickle.load(fopen)
        with open(xgb_pkl, 'rb') as fopen:
            XGB = pickle.load(fopen)
    dmatrix = xgb.DMatrix(BOW_XGB.transform(string))
    results = np.argmax(
        XGB.predict(dmatrix, ntree_limit = XGB.best_ntree_limit), axis = 1
    )
    return [(string[no], entities_labels[i]) for no, i in enumerate(results)]


def get_available_entities_models():
    return ['char', 'word', 'concat']


char_frozen = home + '/entities-char-frozen.h5'
word_frozen = home + '/entities-word-frozen.h5'
concat_frozen = home + '/entities-concat-frozen.h5'

char_settings = home + '/entities-char-settings.json'
word_settings = home + '/entities-word-settings.json'
concat_settings = home + '/entities-concat-settings.json'

char_frozen_location = 'char-bidirectional.h5'
char_settings_location = 'char-bidirectional.json'

word_frozen_location = 'crf-lstm-bidirectional.h5'
word_settings_location = 'crf-lstm-bidirectional.json'

concat_frozen_location = 'crf-lstm-concat-bidirectional.h5'
concat_settings_location = 'crf-lstm-concat-bidirectional.json'


def deep_entities(model = 'concat'):
    assert isinstance(model, str), 'model must be a string'
    if model == 'char':
        if not os.path.isfile(char_settings):
            print('downloading ENTITIES char settings')
            download_file(char_settings_location, char_settings)
        with open(char_settings, 'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(char_frozen):
            print('downloading ENTITIES char frozen model')
            download_file(char_frozen_location, char_frozen)
        model = get_char_bidirectional(nodes['char2idx'], nodes['tag2idx'])
        model.load_weights(char_frozen)
        return CHAR_MODEL(model, nodes)
    elif model == 'word':
        if not os.path.isfile(word_settings):
            print('downloading ENTITIES word settings')
            download_file(word_settings_location, word_settings)
        with open(word_settings, 'r') as fopen:
            nodes = json.loads(fopen.read())
        if not os.path.isfile(word_frozen):
            print('downloading ENTITIES word frozen model')
            download_file(word_frozen_location, word_frozen)
        model = get_crf_lstm_bidirectional(nodes['word2idx'], nodes['tag2idx'])
        model.load_weights(word_frozen)
        return WORD_MODEL(model, nodes)
    elif model == 'concat':
        if not os.path.isfile(concat_settings):
            print('downloading ENTITIES concat settings')
            download_file(concat_settings_location, concat_settings)
        with open(concat_settings, 'r') as fopen:
            nodes = json.loads(fopen.read())
            nodes['idx2word'] = {
                int(k): v for k, v in nodes['idx2word'].items()
            }
        if not os.path.isfile(concat_frozen):
            print('downloading ENTITIES concat frozen model')
            download_file(concat_frozen_location, concat_frozen)
        model = get_crf_lstm_concat_bidirectional(
            nodes['char2idx'], nodes['word2idx'], nodes['tag2idx']
        )
        model.load_weights(concat_frozen)
        return CONCAT_MODEL(model, nodes)
    else:
        raise Exception('model not supported')



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import keras
import numpy as np
import tensorflow as tf
from keras.models import Model, Input
from keras.layers import (
    LSTM,
    Embedding,
    Dense,
    TimeDistributed,
    Dropout,
    Bidirectional,
    Reshape,
    Concatenate,
    Lambda,
)
from keras_contrib.layers import CRF
from keras import backend as K
from keras.backend.tensorflow_backend import set_session
from .text_functions import (
    entities_textcleaning,
    classification_textcleaning,
    char_str_idx,
    generate_char_seq,
    str_idx,
)

set_session(tf.InteractiveSession())


class CLASSIFIER:
    def __init__(self, model, dictionary, labels):
        self._model = model
        self._dictionary = dictionary
        self._labels = labels

    def predict(self, string):
        assert isinstance(string, str), 'input must be a string'
        string = classification_textcleaning(string, True)
        splitted = string.split()
        batch_x = str_idx([string], self._dictionary, len(splitted), UNK = 3)
        probs = self._model.predict(batch_x)
        return {label: probs[0, no] for no, label in enumerate(self._labels)}

    def predict_batch(self, strings):
        assert isinstance(strings, list) and isinstance(
            strings[0], str
        ), 'input must be list of strings'
        strings = [classification_textcleaning(i, True) for i in strings]
        maxlen = max([len(i.split()) for i in strings])
        batch_x = str_idx(strings, self._dictionary, maxlen, UNK = 3)
        probs = self._model.predict(batch_x)
        dicts = []
        for i in range(probs.shape[0]):
            dicts.append(
                {label: probs[i, no] for no, label in enumerate(self._labels)}
            )
        return dicts


class CONCAT_MODEL:
    def __init__(self, model, settings, is_lower = True):
        self._model = model
        self._settings = settings
        self._is_lower = is_lower

    def predict(self, string):
        assert isinstance(string, str), 'input must be a string'
        string = string.lower() if self._is_lower else string
        string = entities_textcleaning(string)
        batch_x = char_str_idx([string], self._settings['word2idx'], 2)
        batch_x_char = generate_char_seq(
            batch_x, self._settings['idx2word'], self._settings['char2idx']
        )
        results = np.argmax(self._model.predict([batch_x, batch_x_char])[0], 1)
        return [
            (string[no], self._settings['idx2tag'][str(i)])
            for no, i in enumerate(results)
        ]


class WORD_MODEL:
    def __init__(self, model, settings, is_lower = True):
        self._model = model
        self._settings = settings
        self._is_lower = is_lower

    def predict(self, string):
        assert isinstance(string, str), 'input must be a string'
        string = string.lower() if self._is_lower else string
        string = entities_textcleaning(string)
        batch_x = char_str_idx([string], self._settings['word2idx'], 2)
        results = np.argmax(self._model.predict(batch_x)[0], 1)
        return [
            (string[no], self._settings['idx2tag'][str(i)])
            for no, i in enumerate(results)
        ]


class CHAR_MODEL:
    def __init__(self, model, settings, is_lower = True):
        self._model = model
        self._settings = settings
        self._is_lower = is_lower

    def predict(self, string):
        assert isinstance(string, str), 'input must be a string'
        string = string.lower() if self._is_lower else string
        string = entities_textcleaning(string)
        batch_x = char_str_idx(string, self._settings['char2idx'], 0)
        results = np.argmax(self._model.predict(batch_x), 1)
        return [
            (string[no], self._settings['idx2tag'][str(i)])
            for no, i in enumerate(results)
        ]


def get_char_bidirectional(char2idx, tag2idx):
    input_word = Input(shape = (None,))
    model = Embedding(
        input_dim = len(char2idx) + 1, output_dim = 128, mask_zero = True
    )(input_word)
    model = Bidirectional(
        LSTM(units = 128, return_sequences = False, recurrent_dropout = 0.1)
    )(model)
    out = Dense(len(tag2idx), activation = 'softmax')(model)
    model = Model(input_word, out)
    return model


def get_crf_lstm_bidirectional(word2idx, tag2idx):
    input_word = Input(shape = (None,))
    model = Embedding(
        input_dim = len(word2idx) + 1, output_dim = 128, mask_zero = True
    )(input_word)
    model = Bidirectional(
        LSTM(units = 64, return_sequences = True, recurrent_dropout = 0.1)
    )(model)
    model = TimeDistributed(Dense(50, activation = 'relu'))(model)
    crf = CRF(len(tag2idx))
    out = crf(model)
    model = Model(input_word, out)
    return model


def get_crf_lstm_concat_bidirectional(char2idx, word2idx, tag2idx):
    input_word = Input(shape = (None,))
    input_char = Input(shape = (None, None))
    model_char = Embedding(input_dim = len(char2idx) + 1, output_dim = 128)(
        input_char
    )
    s = K.shape(model_char)

    def backend_reshape(x):
        return K.reshape(x, (s[0] * s[1], s[2], 128))

    model_char = Lambda(backend_reshape)(model_char)
    model_char = Bidirectional(
        LSTM(units = 50, return_sequences = True, recurrent_dropout = 0.1)
    )(model_char)

    def sliced(x):
        return x[:, -1]

    model_char = Lambda(sliced)(model_char)

    def backend_reshape(x):
        return K.reshape(x, (s[0], s[1], 100))

    model_char = Lambda(backend_reshape)(model_char)
    model_word = Embedding(
        input_dim = len(word2idx) + 1, output_dim = 64, mask_zero = True
    )(input_word)
    concated_word_char = Concatenate(-1)([model_char, model_word])
    model = Bidirectional(
        LSTM(units = 50, return_sequences = True, recurrent_dropout = 0.1)
    )(concated_word_char)
    model = TimeDistributed(Dense(50, activation = 'relu'))(model)
    crf = CRF(len(tag2idx))
    output = crf(model)
    model = Model(inputs = [input_word, input_char], outputs = output)
    return model



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter('ignore')

import tensorflow as tf
from collections import Counter
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.decomposition import LatentDirichletAllocation
from .text_functions import STOPWORDS
from tqdm import tqdm
import numpy as np

BETA = 0.75
ETA = 0.4


class Model:
    def __init__(
        self,
        data,
        unigram_distribution,
        word_vectors,
        doc_weights_init,
        n_topics = 25,
        batch_size = 32,
        lambda_const = 1.0,
        learning_rate = 1e-3,
    ):

        vocab_size, embedding_dim = word_vectors.shape
        n_windows = len(data)
        n_documents = len(np.unique(data[:, 0]))

        self.DOC_INDEX = tf.placeholder(tf.int32, shape = [None])
        self.WORD_INDEX = tf.placeholder(tf.int32, shape = [None])
        self.TARGET = tf.placeholder(
            tf.int32, shape = [None, data.shape[1] - 2]
        )

        doc_ids = data[:, 0]
        unique_docs, counts = np.unique(doc_ids, return_counts = True)
        weights = np.zeros((len(unique_docs)), 'float32')
        for i, j in enumerate(unique_docs):
            weights[j] = 1.0 / np.log(counts[i])
        weights = tf.convert_to_tensor(weights, np.float32)

        unigram_distribution = unigram_distribution / (
            np.sum(unigram_distribution ** BETA) + 1e-5
        )
        unigram_distribution = tf.convert_to_tensor(
            unigram_distribution, np.float32
        )

        self.topic_vectors = tf.get_variable(
            'topic_vectors',
            shape = (n_topics, embedding_dim),
            initializer = tf.orthogonal_initializer(),
        )

        alpha = 1.0 / n_topics
        self.embedding_doc = tf.get_variable(
            'self.embedding_doc',
            shape = doc_weights_init.shape,
            initializer = tf.initializers.constant(doc_weights_init),
            trainable = True,
        )
        self.embedding_word = tf.get_variable(
            'self.embedding_word',
            shape = word_vectors.shape,
            initializer = tf.initializers.constant(word_vectors),
            trainable = False,
        )

        embedded_doc = tf.nn.embedding_lookup(
            self.embedding_doc, self.DOC_INDEX
        )
        embedded_word = tf.nn.embedding_lookup(
            self.embedding_word, self.WORD_INDEX
        )
        doc_probs = tf.nn.softmax(embedded_doc)
        unsqueezed_doc_probs = tf.expand_dims(doc_probs, 2)
        unsqueezed_topic_vectors = tf.expand_dims(self.topic_vectors, 0)
        doc_vectors = tf.reduce_sum(
            unsqueezed_doc_probs * unsqueezed_topic_vectors, 1
        )

        w = tf.nn.embedding_lookup(weights, self.DOC_INDEX)
        w = (w / (tf.reduce_sum(w) + 1e-5)) * batch_size

        dropout_doc = tf.nn.dropout(doc_vectors, 0.5)
        dropout_word = tf.nn.dropout(embedded_word, 0.5)
        context_vectors = dropout_doc + dropout_word
        targets = tf.nn.embedding_lookup(self.embedding_word, self.TARGET)

        unsqueezed_context = tf.expand_dims(context_vectors, 1)
        log_targets = tf.log(
            tf.nn.sigmoid(tf.reduce_sum(targets * unsqueezed_context, 2))
        )
        noise = tf.multinomial(
            [unigram_distribution],
            batch_size * (data.shape[1] - 2) * batch_size,
        )
        noise = tf.reshape(
            noise, [batch_size, (data.shape[1] - 2) * batch_size]
        )

        noise = tf.nn.embedding_lookup(self.embedding_word, noise)
        noise = tf.reshape(
            noise, [batch_size, data.shape[1] - 2, batch_size, embedding_dim]
        )

        unsqueezed_context = tf.expand_dims(
            tf.expand_dims(context_vectors, 1), 1
        )

        sum_log_sampled = tf.log(
            tf.nn.sigmoid(
                tf.negative(tf.reduce_sum(noise * unsqueezed_context, 3))
            )
        )
        sum_log_sampled = tf.reduce_sum(sum_log_sampled, 2)

        neg_loss = log_targets + sum_log_sampled

        self.negative_loss = tf.negative(
            tf.reduce_mean(w * tf.reduce_sum(neg_loss, 1))
        )
        self.dirichlet_loss = tf.reduce_mean(
            w * tf.reduce_sum(tf.nn.softmax(embedded_word), 1)
        )
        self.dirichlet_loss = self.dirichlet_loss * (
            lambda_const * (1.0 - alpha)
        )
        self.cost = self.negative_loss + self.dirichlet_loss
        tvars = tf.trainable_variables()
        grads, _ = tf.clip_by_global_norm(tf.gradients(self.cost, tvars), 5)
        self.optimizer = tf.train.AdamOptimizer(learning_rate).apply_gradients(
            zip(grads, tvars)
        )


def get_windows(doc, hws = 2):
    inside = [
        (w, doc[(i - hws) : i] + doc[(i + 1) : (i + hws + 1)])
        for i, w in enumerate(doc[hws:-hws], hws)
    ]
    beginning = [
        (w, doc[:i] + doc[(i + 1) : (2 * hws + 1)])
        for i, w in enumerate(doc[:hws], 0)
    ]
    end = [
        (w, doc[-(2 * hws + 1) : i] + doc[(i + 1) :])
        for i, w in enumerate(doc[-hws:], len(doc) - hws)
    ]
    return inside + beginning + end


def process_data(corpus, word2idx, min_words = 5):
    texts = []
    for sentence in corpus:
        if len(sentence.split()) > min_words:
            texts.append(sentence)
    corpus = texts
    texts = ' '.join(corpus)
    words = texts.split()
    word2freq = Counter(words)
    word_counts = [count for _, count in word2freq.most_common()]
    tokenized_docs = [(i, doc.split()) for i, doc in enumerate(corpus)]
    _words = set(words)
    encoded_docs = [
        (i, [word2idx[t] if t in word2idx else 0 for t in doc])
        for i, doc in tokenized_docs
    ]
    word_counts = np.array(word_counts)
    word_counts / sum(word_counts)
    data = []
    for index, (_, doc) in enumerate(encoded_docs):
        windows = get_windows(doc)
        data += [[index, w[0]] + w[1] for w in windows]

    return np.array(data, dtype = 'int64'), word_counts


def generate_lda(corpus, n_topics, max_df = 0.95, min_df = 2, temperature = 7):
    tf_vectorizer = CountVectorizer(
        max_df = max_df, min_df = min_df, stop_words = STOPWORDS
    )
    tfreq = tf_vectorizer.fit_transform(corpus)
    lda = LatentDirichletAllocation(
        n_topics = n_topics,
        max_iter = 5,
        learning_method = 'online',
        learning_offset = 50.0,
        random_state = 0,
    ).fit(tfreq)
    corpus_lda = lda.transform(tfreq)
    doc_weights_init = np.copy(corpus_lda)
    doc_weights_init = np.log(doc_weights_init + 1e-4)
    doc_weights_init /= temperature
    return doc_weights_init


def train_lda2vec(
    data,
    unigram_distribution,
    word_vectors,
    doc_weights_init,
    n_topics,
    batch_size = 32,
    epoch = 10,
):
    sess = tf.InteractiveSession()
    model = Model(
        data,
        unigram_distribution,
        word_vectors,
        doc_weights_init,
        n_topics = n_topics,
        batch_size = batch_size,
    )
    sess.run(tf.global_variables_initializer())
    for i in range(epoch):
        pbar = tqdm(
            range(0, (data.shape[0] // batch_size) * batch_size, batch_size),
            desc = 'train minibatch loop',
        )
        for k in pbar:
            batch_doc = data[k : k + batch_size, 0]
            batch_word = data[k : k + batch_size, 1]
            batch_target = data[k : k + batch_size, 2:]
            negative_loss, dirichlet_loss, loss, _ = sess.run(
                [
                    model.negative_loss,
                    model.dirichlet_loss,
                    model.cost,
                    model.optimizer,
                ],
                feed_dict = {
                    model.DOC_INDEX: batch_doc,
                    model.WORD_INDEX: batch_word,
                    model.TARGET: batch_target,
                },
            )
            pbar.set_postfix(
                cost = loss,
                negative_loss = negative_loss,
                dirichlet_loss = dirichlet_loss,
            )
    doc_weights = model.embedding_doc.eval()
    topic_vectors = model.topic_vectors.eval()
    resulted_word_vectors = model.embedding_word.eval()
    similarity = np.matmul(topic_vectors, resulted_word_vectors.T)
    return doc_weights, similarity
